from __future__ import annotations
from typing import List
import json
from hmac import HMAC
from hashlib import sha256

from trezorlib.misc import encrypt_keyvalue
from trezorlib.tools import parse_path
from trezorlib.client import TrezorClient
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from ..crypto import PATH, FILENAME_MESS, decrypt
from ..entry import Entry
from .managers import Manager, DropboxManager, FileManager
from .store_location import StoreLocation
from ..tag import Tag

class Store:
    def __init__(self, client: TrezorClient, manager: Manager):
        self._master_key = self._get_master_key(client)
        self._manager = manager
        self.entries: List[Entry] = []
        self.tags: List[Tag] = []

    @staticmethod
    def _get_master_key(client: TrezorClient) -> str:
        address_n = parse_path(PATH)
        key = "Activate TREZOR Password Manager?"
        value = bytes.fromhex("2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee2d650551248d792eabf628f451200d7f51cb63e46aadcbb1038aacb05e8c8aee")
        return encrypt_keyvalue(client, address_n, key, value).hex()

    @property
    def filename(self) -> str:
        file_key = self._master_key[:len(self._master_key) // 2]
        return HMAC(file_key.encode(), FILENAME_MESS.encode(), sha256).hexdigest() + '.pswd'

    @staticmethod
    def _decrypt_store(encryptedStore: bytes, store: Store) -> str:
        '''Returns JSON representation of the password store'''
        enc_key = store._master_key[len(store._master_key) // 2:]
        return decrypt(enc_key, encryptedStore).decode('utf8')

    @staticmethod
    def _decode_store(json_store: str, store: Store) -> Store:
        '''Returns object representation of the password store'''
        dict = json.loads(json_store)
        tags = {int(key): Tag.load(dict['tags'][key]) for key in dict['tags']}
        store.tags = [tags[key] for key in tags]
        store.entries = [Entry.load(dict['entries'][key], tags) for key in dict['entries']]
        return store

    @staticmethod
    def load(client: TrezorClient) -> Store:
        location = inquirer.select("Where is the password store located at?", choices=[
            Choice(StoreLocation.GoogleDrive, "Google Drive"),
            Choice(StoreLocation.Dropbox, "Dropbox"),
            Choice(StoreLocation.Filepath, "Filepath")
        ]).execute()
        if location == StoreLocation.GoogleDrive:
            print("Google Drive is not yet supported")
            raise Exception("Unsupported option")
        elif location == StoreLocation.Dropbox:
            store = Store(client, DropboxManager())
        elif location == StoreLocation.Filepath:
            store = Store(client, FileManager())
        else:
            raise Exception("Unreachable code")

        json_store = Store._decrypt_store(store._manager.get_password_store(store.filename), store)
        return Store._decode_store(json_store, store)


import json
from typing import List, Union
from urllib.parse import urlparse

from trezorlib.misc import decrypt_keyvalue
from trezorlib.tools import parse_path
from trezorlib.client import TrezorClient

from .crypto import PATH, decrypt
from .tag import Tag

class Entry:
    def __init__(self, url: str) -> None:
        self.url = url
        self.title: str = None
        self.username: str = None
        self.nonce: str = None
        self.password: str = None
        self.safe_note: str = None
        self.tags: List[Tag] = None

        self.decrypted = False
        self.password_cleartext: str = None
        self.safe_note_cleartext: str = None

    @staticmethod
    def emptify(value: Union[str, None]) -> str:
        return value if value else "<empty>"

    @staticmethod
    def hide(value: Union[str, None]) -> str:
        return value if value else "<hidden>"

    def __str__(self) -> str:
        return (
            f"URL: {self.emptify(self.url)}\n"
            f"Title: {self.emptify(self.title)}\n"
            f"Username: {self.emptify(self.username)}\n"
            f"Password: {self.hide(self.password_cleartext)}\n"
            f"Safe Note: {self.hide(self.safe_note_cleartext)}"
        )

    @staticmethod
    def load(dict, tags: dict[str, Tag]):
        entry = Entry(dict["title"]) # title field in JSON is actually the url
        entry.__dict__.update(dict)
        entry.title = dict["note"] # and note field is the title
        entry.tags = [tags[key] for key in dict["tags"]]
        return entry

    def _get_entry_key(self, client: TrezorClient):
        address_n = parse_path(PATH)

        url = urlparse(self.url)
        if url.scheme in ('ftp', 'http', 'https'):
            domain = url.netloc
        else:
            domain = self.url
        key = f'Unlock {domain} for user {self.username}?'
        value = bytes.fromhex(self.nonce)
        return decrypt_keyvalue(client, address_n, key, value, ask_on_encrypt=False).hex()

    def decrypt(self, client: TrezorClient):
        enc_key = self._get_entry_key(client)
        self.password_cleartext = json.loads(decrypt(enc_key, bytes(self.password["data"])).decode("utf8"))
        self.safe_note_cleartext = json.loads(decrypt(enc_key, bytes(self.safe_note["data"])).decode("utf8"))
        self.decrypted = True


from typing import List

from trezorlib.client import get_default_client

from InquirerPy import inquirer

from .entry import Entry
from .store import Store

def select_entry(entries: List[Entry]) -> Entry:
    '''Lets the user select an entry'''

    choices = [{"value": entry, "name":entry.note if entry.note is not None else entry.title} for entry in entries]
    selection = inquirer.fuzzy(
        message="Select an entry:",
        choices=choices,
    ).execute()
    return selection

def print_trezor():
    print("Proceed on your Trezor device")

def cli():
    try:
        client = get_default_client()
    except:
        print("No available Trezor device")
        exit(1)

    try:
        store = Store.load(client)
    except:
        print("Loading of the password store failed")
        exit(1)

    while True:
        try:
            selected_entry = select_entry(store.entries)
            print(selected_entry)

            if not selected_entry.decrypted:
                decrypt = inquirer.confirm(message="Decrypt entry?", default=False).execute()
                if decrypt:
                    try:
                        print_trezor()
                        selected_entry.decrypt(client)
                        print(selected_entry)
                    except:
                        print("Decryption Failed")

            loop = inquirer.confirm(message="Choose another entry?", default=False).execute()
            if not loop:
                break
        except KeyboardInterrupt:
            break

if __name__ == "__main__":
    cli()


from InquirerPy import inquirer
from InquirerPy.validator import PathValidator

from .manager import Manager

class FileManager(Manager):
    def get_password_store(self, filename: str) -> bytes:
        filepath = inquirer.filepath(
            "Enter path to the password store file:",
            default=filename,
            validate=PathValidator(is_file=True, message="Input is not a file")
        ).execute()
        with open(filepath, "rb") as file:
            return file.read()


import dropbox

from .manager import Manager

APP_KEY = "s340kh3l0vla1nv" # APP_KEY of the official TPM, potentionally breaking if maintainers disable PKCE flow for Dropbox auth

class DropboxManager(Manager):
    def __init__(self):
        self.oauth = None

    def authenticate(self):
        auth_flow = dropbox.DropboxOAuth2FlowNoRedirect(APP_KEY, use_pkce=True, token_access_type='offline')
        authorize_url = auth_flow.start()
        print("1. Go to: " + authorize_url)
        print("2. Click \"Allow\" (you might have to log in first).")
        print("3. Copy the authorization code.")
        auth_code = input("Enter the authorization code here: ").strip()
        oauth_result = auth_flow.finish(auth_code)
        self.oauth = oauth_result

    def get_password_store(self, filename: str) -> bytes:
        if not self.oauth:
            self.authenticate()
        with dropbox.Dropbox(oauth2_access_token=self.oauth.access_token) as dbx:
            (_, response) = dbx.files_download("/" + filename)
            return response.content


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="trezor-pass",
    version="0.0.4",
    description="Trezor Password Manager Command-Line interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pilartomas/trezor-password-manager-cli",
    author="Tomas Pilar",
    author_email="thomas7pilar@gmail.com",
    classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)"
    ],
    keywords="trezor, password",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.9, <4",
    install_requires=["trezor", "cryptography", "inquirerpy", "dropbox"],
    entry_points={
        "console_scripts": [
            "trezor-pass=trezorpass.cli:cli",
        ],
    },
    project_urls={
        "Bug Reports": "https://github.com/pilartomas/trezor-password-manager-cli/issues",
    },
)